/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: sock.c,v 1.6 2000/08/26 01:03:48 malekith Exp $ */

#include "bot.h"

#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

int listen_sock, client_sock = -1;
char listen_password[33];
char *detach_info;

#ifdef IPv6
static int setup_socket6()
{
	struct sockaddr_in6 sin;

	listen_sock = socket(AF_INET6, SOCK_STREAM, 0);
	if (listen_sock == -1)
		return -1;
		
	sin.sin6_family = AF_INET6;
	sin.sin6_port = htons(0);
	sin.sin6_addr = in6addr_any;
	
	if (bind(listen_sock, (struct sockaddr *) &sin, sizeof(sin))) {
		close(listen_sock);
		return -1;
	}
	
	if (listen(listen_sock, 128)) {
		close(listen_sock);
		return -1;
	}
	
	return 0;
}
#endif

static int setup_socket4()
{
	struct sockaddr_in sin;

	listen_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (listen_sock == -1) {
		perror("socket");
		exit(1);
	}
	sin.sin_family = AF_INET;
	sin.sin_port = htons(0);
	sin.sin_addr.s_addr = INADDR_ANY;
	bind(listen_sock, (struct sockaddr *) &sin, sizeof(sin));
	listen(listen_sock, 128);
	return 0;
}

static void setup_socket()
{
#ifdef IPv6
	if (setup_socket6())
#endif
		setup_socket4();
}

void set_detach_password(const char *pass)
{
	int i;

	if (pass)
		strncpy(listen_password, pass, 32);
	else
		for (i = 0; i < 32; i++)
			listen_password[i] = arc4random_byte() % 90 + ' ';
	listen_password[32] = 0;
}

void drop_detach_info()
{
	int i, fd;
	char *name;
	struct sockaddr_in sin;
	size_t s;
	struct packet *p;

	for (i = 1;; i++) {
		name = xf("%s/screens/%d", home_dir, i);
		fd = open(name, O_WRONLY | O_CREAT | O_EXCL, 0600);
		if (fd != -1)
			break;
		xfree(name);
	}
	xfree(detach_info);
	detach_info = name;
	s = sizeof(sin);
	
	getsockname(listen_sock, (struct sockaddr*)&sin, &s);
	p = packet_printf("%d%s", ntohs(sin.sin_port), listen_password);
	
	packet_send(fd, 1, p);
	close(fd);
}

void init_listen()
{
	setup_socket();
	set_detach_password(0);
	drop_detach_info();
	
	link_listen_socket();
}

char *show_addr(struct sockaddr *sin, int size)
{
	char buf[1024];
	(void)size;
#ifdef IPv6
	if (size == sizeof(struct sockaddr_in6))
		inet_ntop(AF_INET6,
			  &((struct sockaddr_in6 *) sin)->sin6_addr,
			  buf, 1000);
	else
#endif
		strcpy(buf, 
			inet_ntoa(((struct sockaddr_in *) sin)->sin_addr));

	return xf("%s : %d", buf,
		  ntohs(((struct sockaddr_in *) sin)->sin_port));
}

void quit()
{
	unlink(detach_info);
	close(client_sock);
	exit(0);
}

