/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: parser.c,v 1.15 2000/08/05 22:25:18 malekith Exp $ */

#include "parser.h"

void prtf(const char *win, const char *fmt, ... );
void poops(CTX *ctx,  const char *fmt, ...)
{
	char *s1;
	va_list ap;

	va_start(ap, fmt);
	s1 = vxf(fmt, ap);
	prtf(0, "!!! parse error: %s:%d: %s near ``%s''", ctx->fn, 
	       ctx->line_no, s1, ctx->p);
	xfree(s1);
}

token *make_token(CTX * ctx, int id, char *name, int numb)
{
	token *t;
	
	t = (token *) xmalloc(sizeof(token));
	memset(t, 0, sizeof(token));
	
	t->id = id;
	if (name)
		t->u.name = name;
	else
		t->u.number = numb;

	if (ctx) {
		t->file_idx = ctx->file_idx;
		t->line = ctx->line_no;
	}
	return t;
}

void print_x(token *t)
{
	token *p, *tmp;
	
	if (t == 0)
		return;
		
	switch (t->id) {
	case ID_HASH:
		printf("%s[", t->u.name);
		print_x(t->left);
		printf("]");
		break;
	case ID_REGEXP:
		printf("/regexp/");
		break;
	case ID_DEF_INT:
	case ID_DEF_STRING:
	case ID_DEF_PTR:
	case ID_DEF_IHASH:
	case ID_DEF_PHASH:
	case ID_DEF_SHASH:
		printf("%s %s", id_name(t->id), t->u.name);
		if (t->left) {
			printf(" = ");
			print_x(t->left);
		}
		break;
	case ID_INT:
		printf("%d", t->u.number);
		break;
	case ID_VAR:
		printf("%s", t->u.name);
		break;
	case ID_STRING:
		printf("\"%s\"", t->u.name);
		break;
	case ID_EXPR1:
		switch (t->u.number) {
		case '-':
		case '+':
		case '~':
		case '!':
			printf("%c(", t->u.number);
			print_x(t->left);
			printf(")");
			break;
		case 'M':
			printf("--(");
			print_x(t->left);
			printf(")");
			break;
		case 'P':
			printf("++(");
			print_x(t->left);
			printf(")");
			break;
		case 'm':
			printf("(");
			print_x(t->left);
			printf(")--");
			break;
		case 'p':
			printf("(");
			print_x(t->left);
			printf(")++");
			break;
		}
		break;
	case ID_BREAK:
		printf("break ");
		print_x(t->left);
		break;
	case ID_CONTINUE:
		printf("continue ");
		print_x(t->left);
		break;
	case ID_RETURN:
		printf("return ");
		print_x(t->left);
		break;
	case ID_DO_WHILE:
		printf("do {\n");
		print_x(t->left);
		printf("} while (");
		print_x(t->u.top);
		printf(")");
		break;
	case ID_WHILE:
		printf("while (");
		print_x(t->u.top);
		printf(") {\n");
		print_x(t->left);
		printf(";\n}\n");
		break;
	case ID_WALK:
		printf("walk %s {\n", t->u.name);
		print_x(t->left);
		printf(";\n}\n");
		break;
	case ID_IF:
		printf("if (");
		print_x(t->u.top);
		printf(") {\n");
		print_x(t->left);
		if (t->right) {
		 	printf(";\n} else {\n");
			print_x(t->right);
		}
		printf(";\n}\n");
		break;
	case ID_EXPR2:
		printf("(");
		print_x(t->left);
		printf(") %s (", op2_name(t->u.number));
		print_x(t->right);
		printf(")");
		break;
	case ID_CALL:
		printf("%s(",t->u.name);
		for (p = t->left; p; p = p->next) {
			tmp = p->next;
			p->next = 0;	/* cheat it, so ';' won't be printed */
			print_x(p);
			if((p->next = tmp))
				printf(", ");
		}
		printf(")");
		break;
	}
	if (t->next) {
		printf(";\n");
		print_x(t->next);
	}
}

token *link_end(token ** h, token ** x, token * what)
{
	if ((*h) == 0) {
		(*x) = (*h) = what;
	} else {
		(*x)->next = what;
	}
	if (*x) {
		while ((*x)->next)
			(*x) = (*x)->next;
	}
	return what;
}

static token *get_block(CTX * ctx);
/* get till ; */
token *get_instruction(CTX * ctx)
{
	token *x, *h = 0;

	for (;;) {
		if (get_token(ctx) == 0) {
			poops(ctx, "expression not finished; (EOF)");
			goto shit;
		}
		if (ctx->type == T_OP && ctx->iv == '{' && h == 0) {
			h = get_block(ctx);
			if (ctx->type != T_OP || ctx->iv != '}') {
				poops(ctx, "{ block not closed (EOF)");
				goto shit;
			}
			return h;
		} else if (h == 0 && ctx->type == T_WORD && is_keyword(ctx->sv)) {
			return handle_keyword(ctx, ctx->sv);
		} else
			ctx->leave_it = 1;
	
		if (link_end(&h, &x, get_expression(ctx)) == 0)
			goto shit;
		
		if (ctx->type != T_OP || (ctx->iv!=','&&ctx->iv!=';')) {
			poops(ctx, "expression finished with WRONG OP");
			goto shit;
		}
		if (ctx->iv == ';')
			break;
	}
	return h;
shit:
	kill_token(h);
	return 0;
}

/* get 'till } | EOF */
static token *get_block(CTX * ctx)
{
	token *x, *h = 0;

	for (;;) {
		if (get_token(ctx) == 0)
			break;
		if (ctx->type == T_OP && ctx->iv == '}')
			break;
		ctx->leave_it = 1;
		if (link_end(&h, &x, get_instruction(ctx)) == 0) {
			kill_token(h);
			return 0;
		}
	}
	
	return h ? h : make_token(ctx, ID_INT, 0, 1);
}

byte *parse(CTX * ctx)
{
	token *x;
	
	x = get_block(ctx);
	
	if (get_token(ctx)) {
		poops(ctx, "unbalanced }");
		kill_token(x);
		return 0;
	} else if (x == 0)
		return 0;
	else {
		if (getenv("DUMP_TREE")) {
			print_x(x);
			printf("\n");
		}

		return pack(x, 0, 0, 0);
	}
}
