/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: var.c,v 1.12 2000/08/06 18:41:15 malekith Exp $ */

#include "vm.h"

void drop_on_stack(var *p, frame *f)
{
	p->lval = 1;
	if (p->name == 0)
		abort();
	hash_add(f->stack, p);
}

var *get_var(const char *name, frame * f)
{
	var *p;
	p = hash_find_ex(f->stack, name);
	if (p == 0)
		p = hash_find_ex(f->global->stack, name);
	return p;
}

var *get_var_h(const char *name, frame * f, u32 x)
{
	var *p;
	p = hash_find_ex2(f->stack, name, x);
	if (p == 0)
		p = hash_find_ex2(f->global->stack, name, x);
	return p;
}

static var *new_var(const char *name)
{
	var *p;
	p = (var *) xmalloc(sizeof(var));
	memset(p, 0, sizeof(var));
	if (name == 0)
		name = "CoNsT";
	p->name = xstrdup(name);
	return p;
}

var *new_iv(const char *name, int val)
{
	var *p;
	p = new_var(name);
	p->what = V_INT;
	p->var_int = val;
	return p;
}

var *new_svn(const char *name, int n)
{
	var *p;

	p = new_var(name);
	p->what = V_STRING;
	p->var_string = (char*)xmalloc(n + 1);
	p->var_string_size = n + 1;
	return p;
}

var *new_adv(char *val)
{
	var *p;
	p = new_var(0);
	p->what = V_AD;	/* array dereferance */
	p->var_string = val;
	p->var_string_size = 0;
	return p;
}

var *new_sv(const char *name, const char *val)
{
	var *p;
	p = new_var(name);
	p->what = V_STRING;
	p->var_string = xstrdup(val);
	p->var_string_size = strlen(val) + 1;
	return p;
}

var *new_pv(const char *name, void *val)
{
	var *p;
	p = new_var(name);
	p->what = V_PTR;
	p->var_ptr = val;
	return p;
}

var *copy(const var * src)
{
	var *p;
	
	if (src == 0)
		return 0;
		
	p = new_var(src->name);
	p->what = src->what;
	p->lval = 0;
	switch (src->what) {
	case V_VOID:
		break;
	case V_INT:
		p->var_int = src->var_int;
		break;
	case V_PTR:
		p->var_ptr = src->var_ptr;
		break;
	case V_STRING:
		p->var_string = xstrdup(src->var_string);
		p->var_string_size = strlen(src->var_string) + 1;
		break;
	default:
		abort();
	}
	return p;
}

void free_var(var * v)
{
	kill_var(v);
	xfree(v);
}

void kill_var(var * v)
{
	if (v == 0)
		return;
	xfree(v->name);
	if (v->what == V_STRING)
		xfree(v->var_string);
	v->name = v->var_string = (char*)-1;
}

void set_sv(var *v, const char *val)
{
	int len, n;

	if (val == 0)
		val = "";
	len = strlen(val) + 1;

	if (v->what != V_STRING || v->var_string_size < len) {
		if (v->what == V_STRING)
			xfree(v->var_string);
		v->what = V_STRING;	
		n = 2;
		while (n < len)
			n *= 2;
		v->var_string = (char*)xmalloc(n);
		v->var_string_size = n;
	}
	strcpy(v->var_string, val);
}

void set_iv(var *v, int val)
{
	if (v->what == V_STRING)
		xfree(v->var_string);
	v->what = V_INT;
	v->var_int = val;
}

extern frame global;

var *sv(const char *name)
{
	var *v;

	v = hash_find_ex(global.stack, name);
	if (v == 0) {
		v = new_sv(name, "");
		drop_on_stack(v, &global);
	}

	if (v->what != V_STRING) {
		v->what = V_STRING;
		v->var_string = xstrdup("");
		v->var_string_size = 2;
	}
	return v;
}

var *iv(const char *name)
{
	var *v;

	v = hash_find_ex(global.stack, name);
	if (v == 0) {
		v = new_iv(name, 0);
		drop_on_stack(v, &global);
	}

	if (v->what != V_INT) {
		if (v->what == V_STRING)
			xfree(v->var_string);
		v->what = V_INT;
		v->var_int = 0;
	}
	return v;
}
