/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: vm_hash.c,v 1.5 2000/08/05 22:25:18 malekith Exp $ */

#include "vm.h"

static hash *find_hash(const char *hn, frame * f)
{
	hash *p;
	
	for (p = f->hashes; p; p = p->next)
		if (strcmp(p->name, hn) == 0)
			break;
			
	if (p == 0 && f->global != f)
		return find_hash(hn, f->global);
		
	return p;
}

void hash_add_one(frame * f, const char *name, int type)
{
	hash *h;
	
	if (find_hash(name, f))	/* don't redefine hashes */
		return;
	
	h = hash_new(type);
	h->name = xstrdup(name);
	h->next = f->hashes;
	f->hashes = h;
}

var *hash_entry(vm_token *t, frame *f)
{
	hash *h;
	var *key, *v;
	
	if ((h = find_hash(t->u.name, f)) == 0)
		return 0;

	key = exec(t->left, f);
	
	if (key == 0 || (key->what != V_STRING && key->what != V_INT)) {
		croak(t, "expression didn't result in valid key");
		return 0;
	}
	
	if (key->what == V_INT) {
		key->what = V_STRING;
		key->var_string = xf("%d", key->var_int);
	}
	
	v = hash_find_ex(h, key->var_string);
	
	if (v == 0) {
		switch (h->type) {
		case V_INT:
			v = new_iv(key->var_string, 0);
			break;
		case V_STRING:
			v = new_sv(key->var_string, "");
			break;
		case V_PTR:
			v = new_pv(key->var_string, 0);
			break;
		default:
			abort();
		}
		hash_add(h, v);
		v->lval = 1;
	}

	free_var(key);
	
	return v;
}

void kill_frame(frame * f)
{
	hash *tmp;
	
	while (f->hashes) {
		tmp = f->hashes;
		f->hashes = f->hashes->next;
		hash_kill_one(tmp);
	}
	hash_kill_one(f->stack);
	regexp_info_free(f->regexp);
}

static var *walk1(hash *h, var * key, var * val, char *k, 
		  frame * f, vm_token * t)
{
	var *p;
	
	if ((p = hash_find_ex(h, k)) == 0)
		return 0;
		
	if (key->what == V_STRING)
		xfree(key->var_string);
		
	if (val->what == V_STRING)
		xfree(val->var_string);
	
	key->var_string_size = 0;
	key->var_string = k;
	key->what = V_STRING;
	
	val->var_string_size = 0;
	val->what = p->what;

	if (val->what == V_STRING)
		val->var_string = xstrdup(p->var_string);
	else
		val->var_ptr = p->var_ptr;

	return exec_loop(t->left, f);
}

var *hash_walk(vm_token * t, frame * f)
{
	int i;
	hash *h;
	var *p, *v;
	var *key, *val;
	char **keys, **pk;

	h = find_hash(t->u.name, f);
	if (h == 0) {
		croak(t, "walk on undefined hash `%s' requested",
		      t->u.name);
		return 0;
	}
	v = 0;

	key = hash_find_ex(f->stack, "$key");
	val = hash_find_ex(f->stack, "$val");

	if (key == 0) {
		key = new_iv("$key", 0);
		drop_on_stack(key, f);
	}
	if (val == 0) {
		val = new_iv("$val", 0);
		drop_on_stack(val, f);
	}

	pk = keys =
	    (char **) xmalloc((h->num_entries + 1) * sizeof(char *));

	for (i = 0; i < h->size; i++)
		for (p = h->entries[i]; p; p = p->next)
			*pk++ = xstrdup(p->name);
	*pk = 0;

	if (pk - keys != h->num_entries)
		abort();

	for (pk = keys; *pk; pk++) {
		free_var(v);
		v = walk1(h, key, val, *pk, f, t);

		if (f->continue_lev)
			f->continue_lev--;

		if (f->break_lev) {
			f->break_lev--;
			break;
		}
	}

	xfree(keys);

	return v;
}

int undef_hash(vm_token * t, frame * f)
{
	hash *h;
	var *key;

	if ((h = find_hash(t->u.name, f)) == 0)
		return -1;

	key = exec(t->left, f);

	if (key == 0 || (key->what != V_STRING && key->what != V_INT))
		return -1;

	if (key->what == V_INT) {
		key->what = V_STRING;
		key->var_string = xf("%d", key->var_int);
	}
	
	hash_del(h, key->var_string);

	free_var(key);

	return 0;
}

var *fhv(frame *f, const char *hash_name, const char *key, int type)
{
	hash *h;
	var *v;
	
	h = find_hash(hash_name, f);
	if (h == 0) {
		hash_add_one(f, hash_name, type);
		h = find_hash(hash_name, f);
	}

	v = hash_find_ex(h, key);
	if (v == 0) {
		v = type == V_INT ? new_iv(key, 0) : new_sv(key, "");
		hash_add(h, v);
	}

	return v;
}
