/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: sniffy.cc,v 1.7 2000/08/27 18:32:51 malekith Exp $ */

#include "config.h"

#include <sys/wait.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "sniffy.h"


extern int sock_fd;

int freeze = 0;

char *status_line;
void draw_statusline()
{
	put_line(0, s_height - 3, status_line ? status_line 
			: "\e\007 this is status line");
	
	go(0, s_height - 2);
	int x = 0;
	IrcWindow *w = win_head;
	
	int limit = s_width;
	limit -= get_bool(B_SHORT_STATUS_LINE) ? 9 + 1: 9 + 5;
	while (w && x < limit) {
		int color = get_color(C_STATUS_NORMAL);
		if (w == win_topmost)
			color = get_color(C_STATUS_ACTIVE);
		else if (w->changed) {
			color = get_color(C_STATUS_CHANGED);
		}
		set_color(color);

		if (get_bool(B_SHORT_STATUS_LINE))
			x += print("%d%-.6s ", w->no, w->channel[0] ? 
						w->channel : "[stat]");
		else
			x += print("[%d] %-.8s ", w->no, 
				w->channel[0] ? w->channel : "[status]");
		w = w->next;
	}
	while (x++ < s_width)
		print(' ', get_color(C_STATUS_NORMAL));
}


extern "C" void handle_conf(int cmd, struct packet *p);

void handle_packet(int cmd, struct packet *p)
{
	char *win, *arg;
	
	switch (cmd) {
	case CMD_PRINT:
		packet_scanf(p, "%s%s", &win, &arg);
		send_out(win, arg);
		free(win);
		free(arg);
		break;
	case CMD_EXIT:
		packet_scanf(p, "%s", &arg);
		send_out(0, arg);
		free(arg);
		exit(0);
		break;
	case CMD_NEW_WINDOW:
		packet_scanf(p, "%s", &win);
		add_window(win);
		free(win);
		break;
	case CMD_KILL_WINDOW:
		packet_scanf(p, "%s", &win);
		kill_window(win);
		free(win);
		break;
	case CMD_CONF_COLORS:
	case CMD_CONF_BOOL1:
		handle_conf(cmd, p);
		break;
	case CMD_SET_STATUS_LINE:
		xfree(status_line);
		packet_scanf(p, "%s", &status_line);
		draw_statusline();
		break;
	case CMD_FREEZE:
		packet_scanf(p, "%d", &freeze);
		break;
	}
}

void input_key(const char *key, char *caption, int *pos, int clen)
{
	struct packet *p;
	int cmd;
	char *win, *cap;
	
	if (strcmp(key, "enter") == 0) {
		p = packet_printf("%s%s", win_topmost->channel, caption);
		packet_send(sock_fd, CMD_INPUT, p);
		*caption = 0;
		*pos = 0;
	} else {
		p = packet_printf("%s%s%d%s", win_topmost->channel,
			key, *pos, caption);
		packet_send(sock_fd, CMD_KEY,p);
		xfree(p);
		for (;;) {
			p = packet_recv(sock_fd, &cmd);
			if (p == 0)
				exit(0);
			if (cmd == CMD_SET_INPUT)
				break;
			handle_packet(cmd, p);
			xfree(p);
		}
		packet_scanf(p, "%s%d%s", &win, pos, &cap);
		xfree(win);	/* it should be the same ... */
		strncpy(caption, cap, clen);
	}
	
	xfree(p);
}


void draw_screen()
{
	if (!win_topmost)
		return;
	win_topmost->draw();
	draw_statusline();
	input_line->draw();
}

void global_redraw()
{
	draw_screen();
	f_repaint();
}

// -- main --
extern char *xversion, *xdate;
extern "C" void iface_init(int retach);

int main(int argc, char **argv)
{
	fd_set xx;
	
	printf("schizophrenia %s [%s]\n",xversion,xdate);
	printf("y2k by malekith@pld.org.pl\n");
	printf("connecting...\n");

	(void)argc;
	
	iface_init(!*++argv);
	init_curses();
	add_window("");
	
	for (;;) {
		FD_ZERO(&xx);
		FD_SET(0, &xx);
		FD_SET(sock_fd, &xx);
		select(sock_fd + 1, &xx, 0, 0, 0);
		if (FD_ISSET(sock_fd, &xx)) {
			struct packet *p;
			int cmd;
			
			p = packet_recv(sock_fd, &cmd);
			if (p == 0)
				exit(0);
			handle_packet(cmd, p);
		}
		if (FD_ISSET(0, &xx)) {
			input_line->handle_input();
			input_line->handle_input();
		}
		input_line->draw();
		if (!freeze)
			repaint();
	}
	
	return 0;
}
